import os
import time

import pandas as pd
from scipy import stats


class Syn_Population(object):
    def __init__(
        self,
        location,
        db,
        column_names_config,
        scenario_config,
        run_ipf_obj,
        run_ipu_obj,
        draw_population_obj,
        entities,
        housing_entities,
        person_entities,
    ):

        self.location = location
        self.db = db
        self.column_names_config = column_names_config
        self.scenario_config = scenario_config

        self.run_ipf_obj = run_ipf_obj
        self.geo_constraints = run_ipf_obj.geo_constraints
        self.geo_frequencies = run_ipf_obj.geo_frequencies
        self.region_constraints = run_ipf_obj.region_constraints

        self.run_ipu_obj = run_ipu_obj
        self.geo_row_idx = run_ipu_obj.geo_row_idx
        self.geo_stacked = run_ipu_obj.geo_stacked
        self.region_sample_weights = run_ipu_obj.region_sample_weights

        self.draw_population_obj = draw_population_obj

        self.entities = entities
        self.housing_entities = housing_entities
        self.person_entities = person_entities

        self.geo_name = self.column_names_config.geo
        self.region_name = self.column_names_config.region
        self.hid_name = self.column_names_config.hid
        self.pid_name = self.column_names_config.pid
        self.unique_id_in_geo_name = "unique_id_in_geo"

        self.pop_syn = None
        self.pop_syn_data = {}

        # The following sort_values column variables
        # are useful to separate synthetic population when same
        # sample unit is drawn multiple times
        self.pop_syn_geo_id_columns = [
            self.geo_name, self.unique_id_in_geo_name]
        self.pop_syn_all_id_columns = [
            self.geo_name, self.hid_name, self.unique_id_in_geo_name]
        self.pop_syn_housing_matching_id_columns = [
            self.geo_name, self.hid_name]
        self.pop_syn_person_matching_id_columns = [
            self.geo_name, self.hid_name, self.pid_name]

        self.pop_rows_syn_dict = {}
        self.housing_syn_dict = {}
        self.person_syn_dict = {}
        self.controls = {}
        self.geo_controls = {}
        self.region_controls = {}

        self._create_preliminaries()

    def _create_preliminaries(self):
        self._create_ds()
        self._create_meta_data()
        self._create_prepare_output_directory()

    def _create_ds(self):
        self.housing_stacked_sample = self._get_stacked_sample(
            self.housing_entities)
        self.person_stacked_sample = self._get_stacked_sample(
            self.person_entities)

    def _create_meta_data(self):
        region_controls_config = self.scenario_config.control_variables.region
        geo_controls_config = self.scenario_config.control_variables.geo

        controls_config_list = [geo_controls_config, region_controls_config]
        for entity in self.entities:
            self.controls[entity] = self._return_controls_for_entity(
                controls_config_list, entity)

        controls_config_list = [geo_controls_config]
        for entity in self.entities:
            self.geo_controls[entity] = self._return_controls_for_entity(
                controls_config_list, entity)

        controls_config_list = [region_controls_config]
        for entity in self.entities:
            self.region_controls[entity] = self._return_controls_for_entity(
                controls_config_list, entity)

        self.entity_types_dict = {}
        for entity in self.housing_entities:
            self.entity_types_dict[entity] = "housing"
        for entity in self.person_entities:
            self.entity_types_dict[entity] = "person"

        self.entity_types = ["housing", "person"]

    def _create_prepare_output_directory(self):
        current_time_str = time.strftime("%Y-%m-%d %H-%M-%S", time.localtime())
        foldername = "%s %s" % (
            current_time_str, self.scenario_config.description)
        self.outputlocation = os.path.join(self.location, foldername)
        if not os.path.exists(self.outputlocation):
            os.makedirs(self.outputlocation)

        self.filetype_sep_dict = {"csv": ","}

    def _return_controls_for_entity(self, controls_config_list, entity):
        controls = []
        for controls_config in controls_config_list:
            controls += controls_config[entity].return_list()
        return controls

    def _get_stacked_sample(self, entities):
        # sample_list = []
        sample_list = [self.db.sample[entity] for entity in entities]
        stacked_sample = pd.concat(sample_list).fillna(0)
        # stacked_sample = stacked_sample.loc[stacked_sample['sample_geo'].isin(
        #     self.db.sample_geo_ids)]
        # stacked_sample.sort_values(by='' ,inplace=True)
        return stacked_sample

    def add_records(self):
        geo_id_rows_syn_dict = self.draw_population_obj.geo_id_rows_syn_dict
        for geo_id, geo_id_rows_syn in geo_id_rows_syn_dict.items():
            geo_id_pop_syn = self._get_stacked_geo_for_geo_id(
                geo_id, geo_id_rows_syn)
            self.pop_rows_syn_dict[geo_id] = geo_id_pop_syn
            self.pop_rows_syn_dict[geo_id][self.unique_id_in_geo_name] = list(
                range(1, geo_id_rows_syn.shape[0] + 1))
            #  self.pop_rows_syn_dict[geo_id].head()
            # raw_input()

    def _add_records_for_geo_id(self, geo_id, geo_id_rows_syn):
        # if self.pop_syn is not None:
        #    self.pop_syn = self.pop_syn.append(
        #        self.get_records_for_geo_id(geo_id, geo_id_rows_syn),
        #        ignore_index=True)
        # else:
        #    self.pop_syn = self.get_records_for_geo_id(
        #    geo_id, geo_id_rows_syn)
        geo_id_pop_syn = self._get_stacked_geo_for_geo_id(
            geo_id, geo_id_rows_syn)

        self.pop_rows_syn_dict[geo_id] = geo_id_pop_syn
        # self.housing_syn_dict[geo_id] = (
        #    self.get_housing_data_for_indexes(geo_id_pop_syn))
        # self.person_syn_dict[geo_id] = (
        #    self.get_person_data_for_indexes(geo_id_pop_syn))

    def _get_stacked_geo_for_geo_id(self, geo_id, geo_id_rows_syn):
        geo_id_pop_syn = self.geo_stacked.take(geo_id_rows_syn).copy()
        geo_id_pop_syn[self.geo_name] = geo_id
        #  "\trows", geo_id_pop_syn.shape
        return geo_id_pop_syn

    def _get_housing_data_for_indexes(self, geo_id_pop_syn):
        housing_data = geo_id_pop_syn.loc[:, [self.geo_name]].join(
            self.housing_stacked_sample)
        ("housing rows:", housing_data.shape)
        return housing_data

    def _get_person_data_for_indexes(self, geo_id_pop_syn):
        person_data = geo_id_pop_syn.loc[:, [self.geo_name]].join(
            self.person_stacked_sample)
        ("person rows:", person_data.shape)
        return person_data

    def prepare_data(self):
        self._stack_records()
        self._create_synthetic_population()
        self._create_index()

    def _stack_records(self):
        t = time.time()
        self.pop_syn = pd.concat(
            list(self.pop_rows_syn_dict.values()), copy=False)
        # self.pop_syn_data["housing"] = pd.concat(
        #    self.housing_syn_dict.values(), copy=False)
        # self.pop_syn_data["person"] = pd.concat(
        #    self.person_syn_dict.values(), copy=False)
        #  self.pop_syn_data["housing"].shape
        #  self.pop_syn_data["person"].shape
        ("Time elapsed for stacking population is : %.4f" % (time.time() - t))

    def _create_synthetic_population(self):
        t = time.time()
        self.pop_syn_data["housing"] = self.pop_syn.loc[:, self.pop_syn_geo_id_columns].merge(
            self.housing_stacked_sample, left_index=True, right_index=True
        )
        self.pop_syn_data["person"] = self.pop_syn.loc[:,
                                                       self.pop_syn_geo_id_columns].merge(self.person_stacked_sample,
                                                                                          left_index=True, right_index=True)
        ("\tSize of the housing population table:",
         (self.pop_syn_data["housing"].shape))
        ("\tSize of the person population table:",
         (self.pop_syn_data["person"].shape))
        #  "Time elapsed for synthetic population 1 is : %.4f" % (
        #    time.time() - t)

    def _create_index(self):
        t = time.time()
        # self.pop_syn_data["housing"].reset_index(inplace=True)
        self.pop_syn_data["housing"].set_index(
            self.pop_syn_housing_matching_id_columns, inplace=True, drop=False)
        # self.pop_syn_data["housing"].sort_values(inplace=True)

        # self.pop_syn_data["person"].reset_index(inplace=True)
        self.pop_syn_data["person"].set_index(
            self.pop_syn_person_matching_id_columns, inplace=True, drop=False)
        # self.pop_syn_data["person"].sort_values(inplace=True)
        #  "Time elapsed for index is : %.4f" % (time.time() - t)

    def export_outputs(self):
        ("\tGenerating Outputs")
        t = time.time()
        self._export_performance_data()
        self._export_multiway_tables()
        self._export_summary()
        self._export_weights()
        self._export_synthetic_population()
        self._pretty__scenario_configuration_file_to_output()
        ("\tTime elapsed for generating outputs is : %.4f" % (time.time() - t))

    def _pretty__scenario_configuration_file_to_output(self):
        filepath = os.path.join(self.outputlocation, "%s.yaml" %
                                self.scenario_config.description)
        self.scenario_config.write_to_file(filepath)

    def _export_performance_data(self):
        values_to_export = self.scenario_config.outputs.performance
        #  "Performance values to export:", values_to_export
        if "ipf" in values_to_export:
            self._export_all_df_in_dict(
                self.run_ipf_obj.geo_iters_convergence_dict, "ipf_geo_iters_convergence_")
            self._export_all_df_in_dict(
                self.run_ipf_obj.geo_average_diffs_dict, "ipf_geo_average_diffs_")
            self._export_all_df_in_dict(
                self.run_ipf_obj.region_iters_convergence_dict, "ipf_region_iters_convergence_")
            self._export_all_df_in_dict(
                self.run_ipf_obj.region_average_diffs_dict, "ipf_region_average_diffs_")
        if "reweighting" in values_to_export:
            self._export_df(self.run_ipu_obj.average_diffs,
                            "reweighting_average_diffs")
        if "drawing" in values_to_export:
            self._export_df(
                self.draw_population_obj.draws_performance, "draws")

    def _export_weights(self):
        export_weights_config = self.scenario_config.outputs.weights
        if export_weights_config.export:
            df = pd.DataFrame(self.run_ipu_obj.region_sample_weights)
            if export_weights_config.collate_across_geos:
                df = df.sum(axis=1)
            self._export_df(df, "weights")

    def _export_df(self, df, filename):
        filepath = os.path.join(self.outputlocation, "%s.csv" % filename)
        df.to_csv(filepath)

    def _export_all_df_in_dict(self, dict_of_dfs, fileprefix):
        for key, value in dict_of_dfs.items():
            filename = "%s%s.csv" % (fileprefix, key)
            filepath = os.path.join(self.outputlocation, filename)
            value.to_csv(filepath)

    def _export_multiway_tables(self):
        multiway_tables = self._return_multiway_tables()

        for (filename, filetype), table in multiway_tables.items():
            filepath = os.path.join(self.outputlocation, filename)
            table.to_csv(filepath, sep=self.filetype_sep_dict[filetype])

    def _return_multiway_tables(self):
        multiway_tables = {}
        for table_config in self.scenario_config.outputs.multiway:
            t = time.time()
            (variables, filename, filetype, entity) = (
                table_config.variables.return_list(),
                table_config.filename,
                table_config.filetype,
                table_config.entity,
            )
            entity_type = self.entity_types_dict[entity]
            multiway_table_entity = self._return_aggregate_by_geo(
                variables, entity_type, entity)
            multiway_tables[(filename, filetype)] = multiway_table_entity
            ("\t\tTime elapsed for each table is: %.4f" % (time.time() - t))
        return multiway_tables

    def _export_synthetic_population(self):
        t = time.time()
        synthetic_population_config = self.scenario_config.outputs.synthetic_population
        sort_values_columns = self.pop_syn_all_id_columns
        for entity_type in self.entity_types:
            (filename, filetype) = (
                synthetic_population_config[entity_type].filename,
                synthetic_population_config[entity_type].filetype,
            )
            filepath = os.path.join(self.outputlocation, filename)
            # self.pop_syn_data[entity_type].to_csv(
            #    filepath, sep=self.filetype_sep_dict[filetype], index=False)
            self.pop_syn_data[entity_type].sort_values(
                by=sort_values_columns, inplace=True)
            self.pop_syn_data[entity_type].reset_index(drop=True, inplace=True)
            self.pop_syn_data[entity_type].index.name = "unique_%s_id" % entity_type
            self.pop_syn_data[entity_type].to_csv(
                filepath, sep=self.filetype_sep_dict[filetype])
        ("\tTime to write syn pop files is: %.4f" % (time.time() - t))

    def _return_aggregate_by_geo(self, variables, entity_type, entity):
        if isinstance(variables, str):
            variables = [variables]
        groupby_columns = ["entity", self.geo_name] + variables
        columns_count = len(groupby_columns)
        self.pop_syn_data[entity_type].index.names = [
            None] * len(self.pop_syn_data[entity_type].index.names)
        multiway_table = self.pop_syn_data[entity_type].groupby(
            groupby_columns).size()
        multiway_table_entity = multiway_table[entity].unstack(
            level=list(range(1, columns_count - 1)))
        return multiway_table_entity

    def _export_summary(self):
        t = time.time()
        summary_config = self.scenario_config.outputs.summary
        marginal_geo = self._return_marginal_geo()
        (geo_filename, geo_filetype) = (
            summary_config.geo.filename, summary_config.geo.filetype)
        filepath = os.path.join(self.outputlocation, geo_filename)
        marginal_geo.to_csv(filepath, sep=self.filetype_sep_dict[geo_filetype])
        #  marginal_geo

        marginal_region = self._return_marginal_region(marginal_geo)
        (region_filename, region_filetype) = (
            summary_config.region.filename, summary_config.region.filetype)
        filepath = os.path.join(self.outputlocation, region_filename)
        marginal_region.to_csv(
            filepath, sep=self.filetype_sep_dict[region_filetype])
        ("\tSummary creation took: %.4f" % (time.time() - t))
        #  marginal_region

    def _return_marginal_region(self, marginal_geo):
        region_to_geo = self.db.geo["region_to_geo"]
        marginal_region = region_to_geo.join(
            marginal_geo, on=self.geo_name, how="inner")[marginal_geo.columns]
        marginal_region = marginal_region.reset_index().groupby(self.region_name).sum()

        marginal_region.columns = pd.MultiIndex.from_tuples(
            marginal_region.columns)
        return marginal_region

    def _return_marginal_geo(self):
        marginal_list = []
        for entity in self.entities:
            entity_type = self.entity_types_dict[entity]
            for variable in self.controls[entity]:
                variable_marginal = self._return_aggregate_by_geo(
                    variable, entity_type, entity)
                marginal_list.append(variable_marginal)
        marginal_geo = self._stack_marginal(marginal_list)
        return marginal_geo

    def _stack_marginal(self, marginal_list):
        marginal_T_list = []
        for marginal in marginal_list:
            marginal = marginal.T.copy()
            marginal["name"] = marginal.index.name
            marginal_T_list.append(marginal)
        stacked_marginal = pd.concat(marginal_T_list)
        stacked_marginal.index.name = "categories"
        stacked_marginal.reset_index(inplace=True)
        stacked_marginal.set_index(["name", "categories"], inplace=True)
        # stacked_marginal.sort_values(inplace=True)  # sort_values by row indices
        return stacked_marginal.T

    def _report_summary(self, geo_id_rows_syn, geo_id_frequencies, geo_id_constraints, over_columns=None):
        geo_id_synthetic = self.geo_stacked.take(geo_id_rows_syn).sum()
        geo_id_synthetic = pd.DataFrame(
            geo_id_synthetic, columns=["synthetic_count"])
        geo_id_synthetic["frequency"] = geo_id_frequencies
        geo_id_synthetic["constraint"] = geo_id_constraints
        geo_id_synthetic["diff_constraint"] = geo_id_synthetic["synthetic_count"] - \
            geo_id_synthetic["constraint"]
        geo_id_synthetic["abs_diff_constraint"] = geo_id_synthetic["diff_constraint"].abs(
        )
        geo_id_synthetic["diff_frequency"] = geo_id_synthetic["synthetic_count"] - \
            geo_id_synthetic["frequency"]
        geo_id_synthetic["abs_diff_frequency"] = geo_id_synthetic["diff_frequency"].abs(
        )

        stat, p_value = stats.chisquare(
            geo_id_synthetic["synthetic_count"], geo_id_synthetic["constraint"])
        aad_in_frequencies = (geo_id_synthetic["abs_diff_frequency"]).mean()
        aad_in_constraints = (geo_id_synthetic["abs_diff_constraint"]).mean()
        sad_in_constraints = (geo_id_synthetic["abs_diff_constraint"]).sum()
        sd_in_constraints = (geo_id_synthetic["diff_constraint"]).sum()

        (
            "%.4f, %f, %f, %f, %f, %f"
            % (stat, p_value, aad_in_frequencies, aad_in_constraints, sad_in_constraints, sd_in_constraints)
        )



import yaml


class ConfigError(Exception):
    pass


def wrap_config_value(value):
    """The method is used to wrap YAML elements as Config objects. So the
    YAML properties can be accessed using attribute access.
    E.g. If config object - x for is specificed as the following YAML:

    attribbute1:
        attribute2     : 'Value'

    then attribute access x.attribute1.attribute2 is used to access "Value".
    Also, x.attribute can be used to access the dictionary {attribute: 'value'}
    """
    if isinstance(value, str):
        return value

    try:
        return value + 0
    except TypeError:
        pass

    return Config(value)


class Config(object):
    """The class returns a Config object that can be used to access the
    different YAML elements used to specify the PopGen project.
    """

    def __init__(self, data):
        self._data = data

    def __getattr__(self, key):
        value = self.return_value(key)
        return wrap_config_value(value)

    def __getitem__(self, key):
        value = self.return_value(key)
        return wrap_config_value(value)

    def return_value(self, key):
        try:
            value = self._data[key]
        except KeyError as e:
            raise ConfigError(
                "Key - %s doesn't exist in the YAML configuration" % key)
        return value

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return self._data.__repr__()

    def return_list(self):
        data_list = []
        for i in self._data:
            data_list.append(i)
        return data_list

    def return_dict(self):
        return self._data

    def write_to_file(self, filepath):
        with open(filepath, "w") as outfile:
            outfile.write(yaml.dump(self._data, default_flow_style=False))


if __name__ == "__main__":
    import yaml

    yaml_f = open("../demo/bmc/configuration.yaml", "r")

    config_dict = yaml.load(yaml_f)
    config_obj = Config(config_dict)



import logging
import os
import time

from yaml import Loader, load

from .config import Config
from .data import DB
from .draw import Draw_Population
from .ipf import Run_IPF
from .output import Syn_Population
from .reweighting import Run_Reweighting


class Project(object):
    """This is the primary class to setup and run PopGen projects."""

    def __init__(self, config_loc):
        self.config_loc = config_loc  # Constructor: Set config file location

    def load_project(self):
        self._load_config()  # Load the config file
        self._populate_project_properties()  # Populate Project Properties
        self._load_data()  # Load Data

    def _load_config(self):
        # TODO: validating config file for YAML
        # TODO: validating YAML config file for field types
        # TODO: validating YAML for consistency across fields/config elements
        # Read config file (consider using open)
        config_f = open(self.config_loc, "r")
        # Load YAML configuration
        config_dict = load(config_f, Loader=Loader)
        # Can be used for reading specific objects of config file
        self._config = Config(config_dict)
        # Get column names as dict
        self.column_names_config = self._config.project.inputs.column_names
        self.entities = self._config.project.inputs.entities  # Get entities as list
        # Get housing_entities as list
        self.housing_entities = self._config.project.inputs.housing_entities
        # Get person_entities as list
        self.person_entities = self._config.project.inputs.person_entities

    def _populate_project_properties(self):
        self.name = self._config.project.name  # Get project name
        self.location = os.path.abspath(
            self._config.project.location)  # Get Absolute location

    def _load_data(self):
        self.db = DB(self._config)  #
        self.db.load_data()

    def run_scenarios(self):
        scenarios_config = self._config.project.scenario
        for scenario_config in scenarios_config:
            scenario_obj = Scenario(
                self.location,
                self.entities,
                self.housing_entities,
                self.person_entities,
                self.column_names_config,
                scenario_config,
                self.db,
            )
            scenario_obj.run_scenario()


class Scenario(object):
    def __init__(self, location, entities, housing_entities, person_entities, column_names_config, scenario_config, db):
        self.location = location
        self.entities = entities
        self.housing_entities = housing_entities
        self.person_entities = person_entities
        self.column_names_config = column_names_config
        self.scenario_config = scenario_config
        self.db = db
        self.t = time.time()

    def run_scenario(self):
        self._get_geo_ids()
        self._run_ipf()
        self._run_weighting()
        self._draw_sample()
        self._report_results()

    def _get_geo_ids(self):
        self.db.enumerate_geo_ids_for_scenario(self.scenario_config)
        # self.db.enumerate_sample_geo_ids_for_scenario(self.scenario_config)

    def _run_ipf(self):
        self.run_ipf_obj = Run_IPF(
            self.entities, self.housing_entities, self.column_names_config, self.scenario_config, self.db
        )
        self.run_ipf_obj.run_ipf()
        ("IPF completed in: %.4f" % (time.time() - self.t))

    def _run_weighting(self):
        reweighting_config = self.scenario_config.parameters.reweighting
        # if reweighting_config.procedure == "ipu":
        #     self._run_ipu()
        # def _run_ipu(self):
        self.run_reweighting_obj = Run_Reweighting(
            self.entities, self.column_names_config, self.scenario_config, self.db
        )
        self.run_reweighting_obj.create_ds()
        self.run_reweighting_obj.run_reweighting(
            self.run_ipf_obj.region_constraints, self.run_ipf_obj.geo_constraints)
        ("Reweighting completed in: %.4f" % (time.time() - self.t))

    def _draw_sample(self):
        self.draw_population_obj = Draw_Population(
            self.scenario_config,
            self.db.geo_ids,
            self.run_reweighting_obj.geo_row_idx,
            self.run_ipf_obj.geo_frequencies,
            self.run_ipf_obj.geo_constraints,
            self.run_reweighting_obj.geo_stacked,
            self.run_reweighting_obj.region_sample_weights,
        )
        self.draw_population_obj.draw_population()
        ("Drawing completed in: %.4f" % (time.time() - self.t))

    def _report_results(self):
        self.syn_pop_obj = Syn_Population(
            self.location,
            self.db,
            self.column_names_config,
            self.scenario_config,
            self.run_ipf_obj,
            self.run_reweighting_obj,
            self.draw_population_obj,
            self.entities,
            self.housing_entities,
            self.person_entities,
        )
        self.syn_pop_obj.add_records()
        self.syn_pop_obj.prepare_data()
        self.syn_pop_obj.export_outputs()
        ("Results completed in: %.4f" % (time.time() - self.t))


def popgen_run(project_config):
    logger = logging.getLgger()
    pass


if __name__ == "__main__":
    import time

    from .config import Config
    from .data import DB

    t = time.time()
    p_obj = Project("../tutorials/1_basic_popgen_setup/configuration.yaml")
    p_obj.load_project()
    p_obj.run_scenarios()
    ("Time it took: %.4f" % (time.time() - t))



import time

import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame
from scipy.optimize import fsolve

# from sympy.abc import x as root
# from sympy import solve


# TODO: Move all DS processing to Syn_Population Class
class Reweighting_DS(object):
    def __init__(self):
        pass

    def get_sample_restructure(self, entity, sample: DataFrame, variable_names, hid_name):
        sample["entity"] = entity
        sample.index.name = None
        groupby_columns = [hid_name, "entity"] + variable_names
        columns_count = len(groupby_columns)
        sample_restruct = sample.groupby(groupby_columns).size().unstack(
            level=list(range(1, columns_count))).fillna(0)
        return sample_restruct

    def get_row_idx(self, sample_restruct):
        row_idx = {}
        contrib = {}
        for column in sample_restruct.columns.values.tolist():
            rows = np.where(sample_restruct[column] > 0)[0]
            row_idx[column] = rows
            contrib[column] = np.array(
                sample_restruct[column].values, order="C", dtype=int)
        return (row_idx, contrib)

    def merge_multi(self, df, on, how):
        return self.reset_index().join(df, on=on, how=how).set_index(self.index.names)

    def get_stacked_sample_restruct(self, sample_restruct_list, all_variables):
        if len(sample_restruct_list) == 0:
            return None
        elif len(sample_restruct_list) == 1:
            return sample_restruct_list[0]

        stacked_sample = sample_restruct_list[0]
        for i, sample_restruct in enumerate(sample_restruct_list[1:]):
            len_left_frame_index = len(stacked_sample.columns.values[0])
            len_right_frame_index = len(sample_restruct.columns.values[0])

            if len_left_frame_index >= len_right_frame_index:
                stacked_sample = stacked_sample.join(sample_restruct,
                                                     how="outer").fillna(0)
            else:
                stacked_sample = sample_restruct.join(stacked_sample,
                                                      how="outer").fillna(0)
        stacked_sample.columns = pd.Index(
            stacked_sample.columns, tupleize_cols=False)
        return stacked_sample


class Run_Reweighting(object):
    def __init__(self, entities, column_names_config, scenario_config, db):
        self.entities = entities
        self.column_names_config = column_names_config
        self.scenario_config = scenario_config
        self.db = db
        self.outer_iterations = self.scenario_config.parameters.reweighting.outer_iterations
        self.inner_iterations = self.scenario_config.parameters.reweighting.inner_iterations
        self.archive_performance_frequency = self.scenario_config.parameters.reweighting.archive_performance_frequency
        self.procedure = self.scenario_config.parameters.reweighting.procedure

    def create_ds(self):
        region_controls_config = self.scenario_config.control_variables.region
        (self.region_stacked, self.region_row_idx, self.region_contrib) = self._create_ds_for_resolution(
            region_controls_config
        )
        geo_controls_config = self.scenario_config.control_variables.geo
        (self.geo_stacked, self.geo_row_idx,
         self.geo_contrib) = self._create_ds_for_resolution(geo_controls_config)
        self._create_sample_weights_df()
        self._create_reweighting_performance_df()

    def _create_ds_for_resolution(self, control_variables_config):
        sample_restruct_list = []
        reweighting_ds_obj = Reweighting_DS()

        hid_name = self.column_names_config.hid
        all_variables = []
        for entity in self.entities:
            variable_names = (control_variables_config[entity]).return_list()
            sample = self.db.sample[entity]
            sample_restruct = reweighting_ds_obj.get_sample_restructure(
                entity, sample, variable_names, hid_name)
            sample_restruct_list.append(sample_restruct)
            all_variables += variable_names

        stacked_sample = reweighting_ds_obj.get_stacked_sample_restruct(
            sample_restruct_list, all_variables)
        row_idx, contrib = reweighting_ds_obj.get_row_idx(stacked_sample)
        #  "Sample stacked\n", stacked_sample[:10]
        return (stacked_sample, row_idx, contrib)

    def _create_sample_weights_df(self):
        self.region_sample_weights = pd.DataFrame(
            index=self.region_stacked.index)

    def _create_reweighting_performance_df(self):
        # TODO: In the future change the frequency at which
        # performance measures are stored as a parameter that is
        # specified by the user
        self.iters_to_archive = list(
            range(0, self.outer_iterations, self.archive_performance_frequency))
        self.average_diffs = pd.DataFrame(
            index=self.db.geo_ids, columns=self.iters_to_archive)

    def run_reweighting(self, region_constraints, geo_constraints):
        for region_id in self.db.region_ids:
            (("\t%s for Region: %d" % (self.procedure, region_id)))
            geo_ids = self.db.get_geo_ids_for_region(region_id)
            len_geo_ids = len(geo_ids)
            sample_weights = np.ones(
                (self.region_stacked.shape[0], len_geo_ids), dtype=float, order="C")
            #  "Outer iterations", self.outer_iterations
            for iter in range(self.outer_iterations):
                t = time.time()
                #  "Region: %s and Iter: %s" % (region_id, iter)
                if region_constraints is not None:
                    sample_weights = self._adjust_sample_weights(
                        sample_weights, region_constraints.loc[region_id])
                #  "After region:", sample_weights[:, :4]
                # raw_input("region_done")
                for index, geo_id in enumerate(geo_ids):
                    #  ("Geo: %s " % geo_id)
                    sample_weights[:, index] = self._adjust_sample_weights(
                        sample_weights[:, index], geo_constraints.loc[geo_id], iters=self.inner_iterations, geo=True
                    )
                    #  "After geo:", sample_weights[:, :4]
                    #  ("sample_weights sum: %.6f" % (
                    #    sample_weights[:, index].sum()))
                    if iter in self.iters_to_archive:
                        self._calculate_populate_average_deviation(
                            geo_id, iter, sample_weights[:,
                                                         index], geo_constraints.loc[geo_id]
                        )
                        pass
                    # raw_input("Geo done %s" %geo_id)
                #  ("\t\t\tOne outer iteration complete in %.4f" %
                #       (time.time() - t))
            self._populate_sample_weights(sample_weights, region_id, geo_ids)
            #  self.average_deviations
            ("\tsample_weights sum:", sample_weights.sum())

    def _adjust_sample_weights(self, sample_weights, constraints, iters=1, geo=False):
        if self.procedure == "ipu":
            return self._ipu_adjust_sample_weights(sample_weights, constraints, iters, geo)
        elif self.procedure == "entropy":
            return self._entropy_adjust_sample_weights(sample_weights, constraints, iters, geo)

    def _ipu_adjust_sample_weights(self, sample_weights, constraints, iters=1, geo=False):
        if geo:
            row_idx = self.geo_row_idx
            contrib = self.geo_contrib
        else:
            row_idx = self.region_row_idx
            contrib = self.region_contrib
        # t = time.time()
        sample_weights = np.array(sample_weights, order="C")
        for i in range(iters):
            for column in reversed(constraints.index):
                # TODO: the reversed iteration of list needs to be replaced
                # with a user specified ordering of the constraints
                if geo is False:
                    weighted_sum = (sample_weights.T.dot(
                        contrib[column])).sum()
                else:
                    weighted_sum = sample_weights.dot(contrib[column])

                if weighted_sum == 0:
                    (("""Weighted sum for column %s in iter %d""" """is zero so don't adjust""" % (
                        column, i)))
                    continue

                adjustment = constraints[column] / weighted_sum
                sample_weights[row_idx[column]] *= adjustment

        return sample_weights

    def _entropy_adjust_sample_weights(self, sample_weights, constraints, iters=1, geo=False):
        if geo:
            row_idx = self.geo_row_idx
            contrib = self.geo_contrib
        else:
            row_idx = self.region_row_idx
            contrib = self.region_contrib
            ones_array = np.ones((sample_weights.shape[1]), order="C")

        # t = time.time()
        sample_weights = np.array(sample_weights, order="C")
        for i in range(iters):
            for column in reversed(constraints.index):
                # TODO: the reversed iteration of list needs to be replaced
                # with a user specified ordering of the constraints
                if geo is False:
                    weights_mul_contrib = np.dot(
                        sample_weights, ones_array) * contrib[column]
                else:
                    weights_mul_contrib = sample_weights * contrib[column]

                root = self._find_root(
                    contrib[column], constraints[column], weights_mul_contrib)
                adjustment = root ** contrib[column]
                sample_weights[row_idx[column]] = np.multiply(
                    sample_weights[row_idx[column]
                                   ].T, adjustment[row_idx[column]]
                ).T

        return sample_weights

    def _find_equation(self, contrib, weights_mul_contrib):
        root_power_weight = np.bincount(contrib, weights=weights_mul_contrib)
        root_power = np.array(list(range(contrib.max() + 1)))
        return root_power[1:], root_power_weight[1:]

    def _optimizing_function(self, root, root_power, root_power_weight, constraint):
        function_value = root_power_weight.dot(root ** root_power) - constraint
        return function_value

    def _find_root(self, contrib, constraint, weights_mul_contrib):
        root_power, root_power_weight = self._find_equation(
            contrib, weights_mul_contrib)

        if len(root_power) == 1:
            root = constraint / root_power_weight
        else:
            starting_value = 0.0
            root = fsolve(self._optimizing_function, starting_value, args=(
                root_power, root_power_weight, constraint))
        return root

    def _calculate_populate_average_deviation(self, geo_id, iter, sample_weights, constraints):
        diff_sum = 0

        sample_weights = np.array(sample_weights, order="C")
        for column in constraints.index:
            weighted_sum = sample_weights.dot(self.geo_contrib[column])
            diff_sum += np.abs(weighted_sum -
                               constraints[column]) / constraints[column]
        average_diff = diff_sum / constraints.shape[0]
        self.average_diffs.loc[geo_id, iter] = average_diff

    def _populate_sample_weights(self, sample_weights, region_id, geo_ids):
        for index, geo_id in enumerate(geo_ids):
            # self.region_sample_weights[(region_id,
            #                            geo_id)] = sample_weights[:, index]
            self.region_sample_weights[geo_id] = sample_weights[:, index]

    def _transform_column_index(self):
        multi_index = pd.MultiIndex.from_tuples(
            self.region_sample_weights.columns.values, names=[
                "region_id", "geo_id"]
        )
        self.region_sample_weights.columns = multi_index



import numpy as np
import pandas as pd
from scipy import stats


class Draw_Population(object):
    def __init__(
        self,
        scenario_config,
        geo_ids,
        geo_row_idx,
        geo_frequencies,
        geo_constraints,
        geo_stacked,
        region_sample_weights,
    ):
        self.scenario_config = scenario_config
        self.geo_ids = geo_ids
        self.geo_row_idx = geo_row_idx
        self.geo_frequencies = geo_frequencies
        self.geo_constraints = geo_constraints
        self.geo_stacked = geo_stacked
        self.region_sample_weights = region_sample_weights

        self.iterations = self.scenario_config.parameters.draws.iterations
        self.seed = self.scenario_config.parameters.draws.seed
        self.pvalue_tolerance = self.scenario_config.parameters.draws.pvalue_tolerance
        self.geo_id_rows_syn_dict = {}
        self.performance_columns = ["p_value", "iterations", "chi_sq_stat"]
        self.draws_performance = pd.DataFrame(
            index=self.geo_ids, columns=self.performance_columns)

    def draw_population(self):
        np.random.seed(self.seed)
        for geo_id in self.geo_ids:
            #  "For geo:", geo_id
            geo_sample_weights = self.region_sample_weights.loc[:, geo_id]
            geo_cumulative_weights = self._return_cumulative_probability(
                geo_sample_weights)
            geo_id_frequencies = self.geo_frequencies.loc[geo_id, :]
            geo_id_constraints = self.geo_constraints.loc[geo_id, :]

            if geo_id_frequencies.sum() == 0:
                continue

            p_value_max = -1
            for iter in range(self.iterations):
                #  "Iter is:", iter, self.iterations
                seed = self.seed + iter
                geo_id_rows_syn = self._pick_households(
                    geo_id_frequencies, geo_cumulative_weights)
                stat, p_value = self._measure_match(
                    geo_id_rows_syn, geo_id_constraints)
                if p_value > self.pvalue_tolerance:
                    (p_value_max, geo_id_rows_syn_max, iter_max, stat_max, max_found) = (
                        p_value,
                        geo_id_rows_syn,
                        iter,
                        stat,
                        True,
                    )
                    self.draws_performance.loc[geo_id, self.performance_columns] = (
                        p_value_max, iter, stat_max)
                    break
                elif p_value > p_value_max:
                    (p_value_max, geo_id_rows_syn_max, iter_max, stat_max, max_found) = (
                        p_value,
                        geo_id_rows_syn,
                        iter,
                        stat,
                        False,
                    )
                    self.draws_performance.loc[geo_id, self.performance_columns] = (
                        p_value_max, iter, stat_max)

            #  "Max found:", max_found, geo_id_frequencies.sum()
            #  "Max iter: %d, %f, %f" % (iter_max, p_value_max, stat_max)
            # self.syn_population.add_records_for_geo_id(
            #    geo_id, geo_id_rows_syn_max)
            self.geo_id_rows_syn_dict[geo_id] = geo_id_rows_syn_max
        #  self.draws_performance

    def _return_cumulative_probability(self, geo_sample_weights):
        geo_cumulative_weights = {}

        for column in self.geo_frequencies.columns.values:
            rows = self.geo_row_idx[column]
            weights = geo_sample_weights.take(rows)
            geo_cumulative_weights[column] = (weights / weights.sum()).cumsum()

            #  geo_cumulative_weights[column]
        return geo_cumulative_weights

    def _pick_households(self, geo_id_frequencies, geo_cumulative_weights):
        last = 0
        (geo_id_frequencies.sum())
        rand_numbers = np.random.random(int(geo_id_frequencies.sum()))
        list_rows_syn_subpop = []
        for column in self.geo_frequencies.columns.values:
            rows = self.geo_row_idx[column]
            column_frequency = geo_id_frequencies[column]
            # (last, int(column_frequency))
            column_bins = np.searchsorted(
                geo_cumulative_weights[column], (rand_numbers[int(
                    last): int(last) + int(column_frequency)]), side="right"
            )
            last += column_frequency
            list_rows_syn_subpop.append(rows.take(column_bins))
        geo_id_rows_syn = np.sort(np.concatenate(list_rows_syn_subpop))
        return geo_id_rows_syn

    def _measure_match(self, geo_id_rows_syn, geo_id_constraints, over_columns=None):

        geo_id_constraints.name = "constraint"
        geo_id_synthetic = self.geo_stacked.take(geo_id_rows_syn).sum()
        # (geo_id_synthetic)
        geo_id_synthetic = pd.DataFrame(
            geo_id_synthetic, columns=["synthetic_count"])
        geo_id_synthetic = geo_id_synthetic.join(
            geo_id_constraints, how="inner")
        (geo_id_synthetic["synthetic_count"],
         geo_id_synthetic["constraint"])
        stat, p_value = stats.chisquare(
            geo_id_synthetic["synthetic_count"], geo_id_synthetic["constraint"])
        return stat, p_value



import numpy as np
import pandas as pd


# TODO: Move all DS processing in the Syn_Population Class
class IPF_DS(object):
    def __init__(self, sample, entity, variable_names, variables_count, variables_cats, sample_geo_name=None):
        self.sample = sample
        self.entity = entity
        self.variable_names = variable_names
        self.variables_count = variables_count
        self.variables_cats = variables_cats
        self.sample_geo_name = sample_geo_name

    def get_seed(self):
        groupby_columns = ["entity"] + self.variable_names
        if self.sample_geo_name is not None:
            groupby_columns = [self.sample_geo_name] + groupby_columns
        self.sample["entity"] = self.entity
        seed = self.sample.groupby(groupby_columns).size().astype(float)
        seed.name = "frequency"
        seed = seed.reset_index()
        seed.set_index(keys=groupby_columns, drop=False, inplace=True)
        return seed

    def get_row_idx(self, seed):
        row_idx = {}
        for index, var in enumerate(self.variable_names):
            for cat in self.variables_cats[var]:
                row_idx[(var, cat)] = seed[var].values == cat
        return row_idx


class IPF(object):
    def __init__(
        self,
        seed_all,
        seed,
        idx,
        marginals,
        ipf_config,
        variable_names,
        variables_cats,
        variables_cats_count,
    ):
        self.seed_all = seed_all
        self.seed = seed
        self.idx = idx
        self.marginals = marginals
        self.ipf_config = ipf_config
        self.variable_names = variable_names
        self.variables_cats = variables_cats
        self.variables_cats_count = variables_cats_count
        self.ipf_iters = self.ipf_config.iterations
        self.ipf_tolerance = self.ipf_config.tolerance
        self.zero_marginal_correction = self.ipf_config.zero_marginal_correction
        self.archive_performance_frequency = self.ipf_config.archive_performance_frequency
        self.average_diff_iters = []
        self.iter_convergence = None

    def run_ipf(self):
        self.frequencies = self._correct_zero_cell_issue()
        # self.frequencies = self.seed["frequency"].values
        for c_iter in range(self.ipf_iters):
            #  "Iter:", c_iter
            self._adjust_cell_frequencies()
            # Checks for convergence every 5 iterations

            # TODO: In the future change the frequency at which
            # performance measures are stored as a parameter that is
            # specified by the user

            if (c_iter % self.archive_performance_frequency) == 0:
                if self._check_convergence():
                    #  "\t\t\tConvergence achieved in %d iter" % (c_iter)
                    self.iter_convergence = c_iter
                    break

    def _correct_zero_cell_issue(self):
        if self.seed.shape[0] != self.seed_all.shape[0]:
            self.seed_all["prob"] = self.seed["frequency"] / \
                self.seed["frequency"].sum()
            null_rows = self.seed_all["prob"].isnull()
            self.seed_all["prob_all"] = self.seed_all["frequency"] / \
                self.seed_all["frequency"].sum()
            self.seed_all.loc[null_rows,
                              "prob"] = self.seed_all.loc[null_rows, "prob_all"]
            borrowed_sum = self.seed_all.loc[null_rows, "prob"].sum()
            adjustment = 1 - borrowed_sum
            self.seed_all.loc[~null_rows, "prob"] *= adjustment

            return self.seed_all["prob"].copy().values
        else:
            return self.seed["frequency"].copy().values

    def _adjust_cell_frequencies(self):
        for var in self.variable_names:
            for cat in self.variables_cats[var]:
                row_subset = self.idx[(var, cat)]
                # TODO: There is a possible issue with how Pandas reads csv
                # files with headers. In the notebook implementation the
                # multiindex is read as alphanumeric whereas in this
                # implementation it is being read as a "alpha" only
                # the below indexing for marginals is just a hack ... need
                # to replace
                marginal = self.marginals.loc[(var, "%s" % cat)]
                if marginal == 0:
                    marginal = self.zero_marginal_correction

                adjustment = marginal / self.frequencies[row_subset].sum()
                self.frequencies[row_subset] *= adjustment

                if (self.frequencies == 0).any():
                    cells_zero_values = self.frequencies == 0
                    self.frequencies[cells_zero_values] = np.finfo(
                        np.float64).tiny

    def _check_convergence(self):
        average_diff = self._calculate_average_deviation()
        self.average_diff_iters.append(average_diff)
        if len(self.average_diff_iters) > 1:
            if np.abs(self.average_diff_iters[-1] - self.average_diff_iters[-2]) < self.ipf_tolerance:
                return True
        return False

    def __marginals(self):
        for var in self.variable_names:
            for cat in self.variables_cats[var]:
                row_subset = self.idx[(var, cat)]
                adjusted_frequency = self.frequencies[row_subset].sum()
                original_frequency = self.marginals.loc[(var, "%s" % cat)]
                (("\t", (var, "%s" % cat), original_frequency, adjusted_frequency))

    def _calculate_average_deviation(self):
        diff_sum = 0
        for var in self.variable_names[:-1]:
            for cat in self.variables_cats[var]:
                row_subset = self.idx[(var, cat)]
                adjusted_frequency = self.frequencies[row_subset].sum()
                # TODO: See above to-do same fix here
                original_frequency = self.marginals.loc[(var, "%s" % cat)]
                if original_frequency == 0:
                    original_frequency = self.zero_marginal_correction
            diff_sum += np.abs(adjusted_frequency -
                               original_frequency) / original_frequency
        average_diff = diff_sum / self.variables_cats_count
        #  "Average Deviation", average_diff
        return average_diff


class Run_IPF(object):
    def __init__(self, entities, housing_entities, column_names_config, scenario_config, db):
        self.entities = entities
        self.housing_entities = housing_entities
        self.column_names_config = column_names_config
        self.scenario_config = scenario_config
        self.db = db
        self.ipf_config = self.scenario_config.parameters.ipf
        self.ipf_rounding = self.ipf_config.rounding_procedure
        self.sample_geo_name = self.column_names_config.sample_geo

    def run_ipf(self):
        region_marginals = self.db.region_marginals
        region_controls_config = self.scenario_config.control_variables.region
        region_ids = self.db.region_ids
        region_to_sample = self.db.geo["region_to_sample"]
        (
            self.region_constraints,
            self.region_constraints_dict,
            self.region_iters_convergence_dict,
            self.region_average_diffs_dict,
        ) = self._run_ipf_for_resolution(region_marginals, region_controls_config, region_ids, region_to_sample)
        self.region_columns_dict = self._get_columns_constraints_dict(
            self.region_constraints_dict)

        geo_marginals = self.db.geo_marginals
        geo_controls_config = self.scenario_config.control_variables.geo
        geo_ids = self.db.geo_ids
        geo_to_sample = self.db.geo["geo_to_sample"]
        (
            self.geo_constraints,
            self.geo_constraints_dict,
            self.geo_iters_convergence_dict,
            self.geo_average_diffs_dict,
        ) = self._run_ipf_for_resolution(geo_marginals, geo_controls_config, geo_ids, geo_to_sample)
        self.geo_columns_dict = self._get_columns_constraints_dict(
            self.geo_constraints_dict)

        if self.ipf_rounding == "bucket":
            self.geo_frequencies = self._get_frequencies_for_resolution(
                geo_ids, self.geo_constraints_dict, "bucket")

    def _run_ipf_for_resolution(self, marginals_at_resolution, control_variables_config, geo_ids, geo_corr_to_sample):
        constraints_list = []
        constraints_dict = {}
        iters_convergence_dict = {}
        average_diffs_dict = {}
        for entity in self.entities:
            (("\tIPF for Entity: %s complete" % entity))

            sample = self.db.sample[entity]
            marginals = marginals_at_resolution[entity]
            variable_names = control_variables_config[entity].return_list()

            if len(variable_names) == 0:
                continue

            variables_cats = self.db.return_variables_cats(
                entity, variable_names)
            variables_count = len(variable_names)
            variables_cats_count = sum(
                [len(cats) for cats in list(variables_cats.values())])
            (seed_geo, seed_all, row_idx) = self._create_ds_for_resolution_entity(
                sample, entity, variable_names, variables_count, variables_cats, self.sample_geo_name
            )

            (constraints, iters_convergence, average_diffs) = self._run_ipf_all_geos(
                entity,
                seed_geo,
                seed_all,
                row_idx,
                marginals,
                variable_names,
                variables_count,
                variables_cats,
                variables_cats_count,
                geo_ids,
                geo_corr_to_sample,
            )
            constraints_dict[entity] = constraints
            iters_convergence[entity] = iters_convergence
            average_diffs_dict[entity] = average_diffs
            constraints_list.append(constraints)
        constraints_resolution = self._get_stacked_constraints(
            constraints_list)
        return (constraints_resolution, constraints_dict, iters_convergence_dict, average_diffs_dict)

    def _create_ds_for_resolution_entity(
        self, sample, entity, variable_names, variables_count, variables_cats, sample_geo_names
    ):
        ipf_ds_geo = IPF_DS(sample, entity, variable_names,
                            variables_count, variables_cats, sample_geo_names)
        seed_geo = ipf_ds_geo.get_seed()

        ipf_ds_all = IPF_DS(sample, entity, variable_names,
                            variables_count, variables_cats)

        seed_all = ipf_ds_all.get_seed()
        row_idx = ipf_ds_all.get_row_idx(seed_all)
        return (seed_geo, seed_all, row_idx)

    def _run_ipf_all_geos(
        self,
        entity,
        seed_geo,
        seed_all,
        row_idx,
        marginals,
        variable_names,
        variables_count,
        variables_cats,
        variables_cats_count,
        geo_ids,
        geo_corr_to_sample,
    ):
        ipf_results = pd.DataFrame(index=seed_all.index)
        ipf_iters_convergence = pd.DataFrame(index=["iterations"])
        ipf_avgerage_diffs = pd.DataFrame(index=["average_percent_deviation"])
        for geo_id in geo_ids:
            #  "\tIPF for Geo: %s for Entity: %s" % (geo_id, entity)
            sample_geo_id = geo_corr_to_sample.loc[geo_id,
                                                   self.sample_geo_name]
            if isinstance(sample_geo_id, pd.Series):
                seed_geo_levels_list = list(range(len(seed_geo.index.names)))
                seed_for_geo_id = seed_geo.loc[sample_geo_id.tolist()].sum(
                    level=seed_geo_levels_list[1:])
                #  (seed_geo.loc[sample_geo_id.tolist()])
                #  sample_geo_id.tolist(), "Satisfied series check"
            # if sample_geo_id.shape[0] >= 1:
            elif sample_geo_id > 0:
                seed_for_geo_id = seed_geo.loc[sample_geo_id.tolist()]
                #  sample_geo_id.tolist(), "Satisfied valid value check"
            elif sample_geo_id == -1:
                seed_for_geo_id = seed_all.copy()
                #  sample_geo_id.tolist(), "Satisfied default value check"
            else:
                raise Exception("Not series nor is it default value of -1")
            #  seed_for_geo_id
            #  "This is the sample geo id"
            # raw_input()

            marginals_geo = marginals.loc[geo_id]
            ipf_obj_geo = IPF(
                seed_all,
                seed_for_geo_id,
                row_idx,
                marginals_geo,
                self.ipf_config,
                variable_names,
                variables_cats,
                variables_cats_count,
            )
            # ipf_obj_geo.correct_zero_cell_issue()
            # ipf_results_geo = ipf_obj_geo.run_ipf()
            ipf_obj_geo.run_ipf()
            ipf_results[geo_id] = ipf_obj_geo.frequencies
            ipf_iters_convergence[geo_id] = ipf_obj_geo.iter_convergence
            ipf_avgerage_diffs[geo_id] = ipf_obj_geo.average_diff_iters[-1]
            #  ('\t', ipf_obj_geo.iter_convergence,
            #         ipf_obj_geo.average_diff_iters)
            if (ipf_results[geo_id] == 0).any():
                raise Exception(
                    """IPF cell values of zero are returned. """ """Needs troubleshooting""")

            # ipf_results[geo_id] = ipf_results_geo["frequency"]
            # raw_input("IPF for Geo: %s for Entity: %s complete"
            #          % (geo_id, self.entity))
        #  ipf_iters_convergence.T
        #  ipf_avgerage_diffs.T
        # raw_input("IPF Results")
        return (ipf_results, ipf_iters_convergence.T, ipf_avgerage_diffs.T)

    def _get_stacked_constraints(self, constraints_list):
        if len(constraints_list) == 0:
            return None
        elif len(constraints_list) == 1:
            return constraints_list[0].T
        stacked_constraints = constraints_list[0].T
        for constraint in constraints_list[1:]:
            len_left_frame_index = len(stacked_constraints.columns.values[0])
            len_right_frame_index = len(constraint.T.columns.values[0])

            if len_left_frame_index >= len_right_frame_index:
                stacked_constraints = stacked_constraints.join(constraint.T)
            else:
                stacked_constraints = constraint.T.join(stacked_constraints)
        stacked_constraints.sort_index(axis=1, inplace=True)

        stacked_constraints.columns = pd.Index(
            stacked_constraints.columns, tupleize_cols=False)
        return stacked_constraints

    def _get_columns_constraints_dict(self, constraints_dict):
        columns_constraints_dict = {}
        for entity, constraints in constraints_dict.items():
            columns_constraints_dict[entity] = constraints.index.values.tolist(
            )
        #  columns_constraints_dict
        return columns_constraints_dict

    def _get_frequencies_for_resolution(self, geo_ids, constraints_dict, procedure="bucket"):
        # TODO: Implemente other procedures for integerizing multiway freq
        frequencies_list = []

        for entity in self.housing_entities:
            (("\tRounding frequencies for Entity: %s complete" % entity))

            frequencies = constraints_dict[entity].copy()

            for geo_id in geo_ids:
                #  ("Rounding Frequencies for Geo: %s for Entity: %s"
                #       % (geo_id, entity))

                frequency_geo = frequencies.loc[:, geo_id].values
                adjusted_frequency_geo = []
                accumulated_difference = 0

                for frequency in frequency_geo:
                    frequency_int = np.floor(frequency)
                    frequency_dec = frequency - frequency_int
                    accumulated_difference += frequency_dec
                    adjustment = accumulated_difference.round()
                    adjusted_frequency_geo.append(frequency_int + adjustment)
                    accumulated_difference -= adjustment
                frequencies.loc[:, geo_id] = adjusted_frequency_geo

            frequencies_list.append(frequencies)
        frequencies_resolution = self._get_stacked_constraints(
            frequencies_list)
        return frequencies_resolution



import os

import numpy as np
import pandas as pd

from .config import ConfigError


class DB(object):
    """This class returns a Inputs object that can be used to handle all
    the Popgen input data files. All input files are stored as Pandas
    dataframe objects. The object maintains the following PopGen inputs:
    - Two geographic correspondence mapping inputs
    1) a mapping between region and geo
    2) a mapping between geo and sample_geo
    - Two sets of sample data files
    1) for housing entities
    2) for person entities
    - Two sets of marginal data files
    1) for housing entities
    2) for person entities
    """

    def __init__(self, config):
        self.config = config  # Get entire config file as a dict
        self.sample = {}  # Get sample data
        self.geo_marginals = {}
        self.region_marginals = {}
        self.geo = {}
        self.geo_ids = None
        self.region_ids = None
        self.sample_geo_ids = None
        self._inputs_config = self.config.project.inputs
        self.location = os.path.abspath(self.config.project.location)

    def load_data(self):

        geo_corr_mapping_config = self._inputs_config.location.geo_corr_mapping
        self.geo = self.get_data(geo_corr_mapping_config)

        sample_config = self._inputs_config.location.sample
        self.sample = self.get_data(sample_config)
        geo_marginals_config = self._inputs_config.location.marginals.geo
        self.geo_marginals = self.get_data(geo_marginals_config, header=[0, 1])

        region_marginals_config = self._inputs_config.location.marginals.region
        self.region_marginals = self.get_data(
            region_marginals_config, header=[0, 1])

        self._enumerate_geo_ids()

    def get_data(self, config, header=0):
        config_dict = config.return_dict()
        data_dict = {}
        for item in config_dict:
            filename = config_dict[item]
            full_location = os.path.join(self.location, filename)
            data_dict[item] = pd.read_csv(
                full_location, index_col=0, header=header)
            data_dict[item].loc[:,
                                data_dict[item].index.name] = data_dict[item].index.values
        return data_dict

    def _enumerate_geo_ids(self):
        geo_to_sample = self.geo["geo_to_sample"]
        self.geo_ids_all = geo_to_sample.index.tolist()
        self.sample_geo_ids = np.unique(geo_to_sample[self._inputs_config
                                                      .column_names
                                                      .sample_geo].values)
        region_to_geo = self.geo["region_to_geo"]
        self.region_ids_all = np.unique(region_to_geo.index.values).tolist()

        # region_to_sample = self.geo["region_to_sample"]
        # self.region_ids = np.unique(region_to_sample.index.values)

    def get_geo_ids_for_region(self, region_id):
        geo_name = self._inputs_config.column_names.geo

        geoids = self.geo["region_to_geo"].loc[region_id,
                                               geo_name].copy().tolist()
        if isinstance(geoids, list):
            return geoids
        else:
            return [geoids]
        # return (self.geo["region_to_geo"].loc[region_id, geo_name].copy().tolist())

    def enumerate_geo_ids_for_scenario(self, scenario_config):
        try:
            self.region_ids = scenario_config.geos_to_synthesize.region.ids
            self.geo_ids = []
            for region_id in self.region_ids:
                self.geo_ids += self.get_geo_ids_for_region(region_id)
        except ConfigError as e:
            print("KeyError", e)
            self.geo_ids = self.geo_ids_all
            # self.sample_geo_ids = self.sample_geo_ids_all
            self.region_ids = self.region_ids_all

    def get_sample_geo_ids_for_region(self, region_id):
        sample_geo_name = self._inputs_config.column_names.sample_geo

        samplegeoids = self.geo["region_to_sample"].loc[region_id,
                                                        sample_geo_name].copy().tolist()
        if isinstance(samplegeoids, list):
            return samplegeoids
        else:
            return [samplegeoids]
        # return (self.geo["region_to_geo"].loc[region_id, geo_name].copy().tolist())

    def enumerate_sample_geo_ids_for_scenario(self, scenario_config):
        try:
            self.region_ids = scenario_config.geos_to_synthesize.region.ids
            self.sample_geo_ids = []
            for region_id in self.region_ids:
                self.sample_geo_ids += self.get_sample_geo_ids_for_region(
                    region_id)
        except ConfigError as e:
            print("KeyError", e)
            # self.geo_ids = self.geo_ids_all
            # # self.sample_geo_ids = self.sample_geo_ids_all
            # self.region_ids = self.region_ids_all

    def return_variables_cats(self, entity, variable_names):
        variables_cats = {}
        for variable_name in variable_names:
            variables_cats[variable_name] = self.return_variable_cats(
                entity, variable_name)
        return variables_cats

    def return_variable_cats(self, entity, variable_name):
        return np.unique(self.sample[entity][variable_name].values).tolist()

    def check_data(self):
        self.check_sample_marginals_consistency()
        self.check_marginals()

    def check(self):
        # TODO: check if the ids entered are consistent with the region ids
        # TODO: check consistency in variables across files
        # TODO: check consistency in categories across files
        # TODO: check consistency in marginals across
        # TODO: check geo ids, sample geo ids, region ids across files
        pass
